/***********************************************************************

                      PARANOIA BLOCK CIPHER

                   256 bit block and 512 bit key

                 Implemented by Alexander Myasnikov

                   Web: www.darksoftware.narod.ru

***********************************************************************/


#include <string.h>
#include <stdlib.h>


typedef signed long s32;
typedef unsigned long u32;
typedef unsigned char u8;

#define M_Q 1.6180339887498949

#define countof(v) (sizeof(v) / sizeof((v)[0]))

#define ROTL(x, u) (_lrotl((x), (u)))
#define ROTR(x, u) (_lrotr((x), (u)))

#define LOAD(p) ( \
        (((u32) (((const u8*)(p))[0])) << 000) \
|       (((u32) (((const u8*)(p))[1])) << 010) \
|       (((u32) (((const u8*)(p))[2])) << 020) \
|       (((u32) (((const u8*)(p))[3])) << 030) \
)
#define SAVE(p, v) do { \
        u8 *_p = (u8*)(p); \
        u32 _v = (u32)(v); \
        _p[0] = (_v >> 000) & 0xff; \
        _p[1] = (_v >> 010) & 0xff; \
        _p[2] = (_v >> 020) & 0xff; \
        _p[3] = (_v >> 030) & 0xff; \
} while (0)





/*---------------------------------------------------------------------------*/

#define PW (((u32) ((3.14159265358979323846 - 3.0) * 4294967295U)) | 1)

#define VALCNT 8
#define BLKSIZ (VALCNT*4)
#define ROUNDS 6
#define KEYCNT (VALCNT * ROUNDS)

struct keysched_s
{
  u32 t[KEYCNT];
  u32 w[VALCNT * 2];
};
typedef struct keysched_s keysched;

keysched key;


void __stdcall __export
setup (u8 * p)
{
  u32 x = 64;
  size_t i;
  u32 l[64];


  /* initialize l */
  for (i = 0; i != countof (l); ++i)
    l[i] = 0;

  for (i = 0; i != x; ++i)
    l[i / 4] |= (p[i] << (010 * (i % 4)));

  /* avalanche addition */
  for (i = 0; i != 3 * countof (key.t); ++i)
    {
      l[i % countof (l)] += PW
	+ l[(i + 3) % countof (l)]
	+ ROTL (l[(i + 11) % countof (l)], 3)
	+ ROTL (l[(i + 31) % countof (l)], 15)
	+ ROTL (l[(i + 53) % countof (l)], 13)
	+ ROTL (l[(i + 57) % countof (l)], 7)
	+ ROTL (l[(i + 62) % countof (l)], 1);
    }

  /* copy results */
  for (i = 0; i != countof (key.t); ++i)
    {
      key.t[i] = l[i];
    }
  for (i = 0; i != countof (key.w); ++i)
    {
      key.w[i] = l[48 + i];
    }
}

static const u32 brt1_v[3] = {
  0xffffffff,
  0x55555555,
  0x00000000
};
static const u32 brt2_v[5] = {
  0xffffffff,
  0x77777777,
  0x33333333,
  0x11111111,
  0x00000000
};
static const u32 brt3_v[9] = {
  0xffffffff,
  0x7f7f7f7f,
  0x3f3f3f3f,
  0x1f1f1f1f,
  0x0f0f0f0f,
  0x07070707,
  0x03030303,
  0x01010101,
  0x00000000
};
static const u32 brt4_v[17] = {
  0xffffffff,
  0x7fff7fff,
  0x3fff3fff,
  0x1fff1fff,
  0x0fff0fff,
  0x07ff07ff,
  0x03ff03ff,
  0x01ff01ff,
  0x00ff00ff,
  0x007f007f,
  0x003f003f,
  0x001f001f,
  0x000f000f,
  0x00070007,
  0x00030003,
  0x00010001,
  0x00000000
};

#define BRTL(x, u, v) ( \
        (((x) & (v)[(u) & (countof(v)-2)]) << ((u) & (countof(v)-2))) \
        | (((x) >> ((countof(v)-1) - ((u) & (countof(v)-2)))) \
                & (v)[(countof(v)-1) - ((u) & (countof(v)-2))]))
#define BRTR(x, u, v) ( \
        (((x) >> ((u) & (countof(v)-2))) & (v)[((u) & (countof(v)-2))]) \
        | (((x) & (v)[(countof(v)-1) - ((u) & (countof(v)-2))]) \
                << ((countof(v)-1) - ((u) & (countof(v)-2)))))

#define BRTL1(x, u) BRTL(x, u, brt1_v)
#define BRTR1(x, u) BRTR(x, u, brt1_v)
#define BRTL2(x, u) BRTL(x, u, brt2_v)
#define BRTR2(x, u) BRTR(x, u, brt2_v)
#define BRTL3(x, u) BRTL(x, u, brt3_v)
#define BRTR3(x, u) BRTR(x, u, brt3_v)
#define BRTL4(x, u) BRTL(x, u, brt4_v)
#define BRTR4(x, u) BRTR(x, u, brt4_v)

void __stdcall __export
crypt (const void *iblock)
{
  register u32 u;
  u32 a, b, c, d, e, f, g, h;
  unsigned i, j;
  u8 *ip = (u8 *) iblock;


  a = LOAD (ip + 0x00);
  b = LOAD (ip + 0x04);
  c = LOAD (ip + 0x08);
  d = LOAD (ip + 0x0c);
  e = LOAD (ip + 0x10);
  f = LOAD (ip + 0x14);
  g = LOAD (ip + 0x18);
  h = LOAD (ip + 0x1c);
  a ^= key.w[0x0];
  b ^= key.w[0x1];
  c ^= key.w[0x2];
  d ^= key.w[0x3];
  e ^= key.w[0x4];
  f ^= key.w[0x5];
  g ^= key.w[0x6];
  h ^= key.w[0x7];

  i = 0;
  j = KEYCNT;
  do
    {
      --j;
      u = key.t[j] + h;
      a += PW;
      a += b;
      a = BRTL3 (a, u >> 26);
      a ^= c;
      a = ROTL (a, u >> 10);
      a = BRTL4 (a, u >> 19);
      a += d;
      a = BRTL2 (a, u >> 29);
      a ^= e;
      a = ROTL (a, u >> 5);
      a = BRTL3 (a, u >> 23);
      a += f;
      a = BRTL4 (a, u >> 15);
      a ^= g;
      a = ROTL (a, u);
      a = BRTL1 (a, u >> 31);
      a += key.t[i];
      ++i;
      u = a;
      a = b;
      b = c;
      c = d;
      d = e;
      e = f;
      f = g;
      g = h;
      h = u;
    }
  while (j != 0);


  a ^= key.w[0x8];
  b ^= key.w[0x9];
  c ^= key.w[0xa];
  d ^= key.w[0xb];
  e ^= key.w[0xc];
  f ^= key.w[0xd];
  g ^= key.w[0xe];
  h ^= key.w[0xf];
  SAVE (ip + 0x00, a);
  SAVE (ip + 0x04, b);
  SAVE (ip + 0x08, c);
  SAVE (ip + 0x0c, d);
  SAVE (ip + 0x10, e);
  SAVE (ip + 0x14, f);
  SAVE (ip + 0x18, g);
  SAVE (ip + 0x1c, h);
}

void __stdcall __export
decrypt (const void *iblock)
{
  register u32 u;
  u32 a, b, c, d, e, f, g, h;
  unsigned i, j;
  u8 *ip = (u8 *) iblock;

  a = LOAD (ip + 0x00);
  b = LOAD (ip + 0x04);
  c = LOAD (ip + 0x08);
  d = LOAD (ip + 0x0c);
  e = LOAD (ip + 0x10);
  f = LOAD (ip + 0x14);
  g = LOAD (ip + 0x18);
  h = LOAD (ip + 0x1c);
  a ^= key.w[0x8];
  b ^= key.w[0x9];
  c ^= key.w[0xa];
  d ^= key.w[0xb];
  e ^= key.w[0xc];
  f ^= key.w[0xd];
  g ^= key.w[0xe];
  h ^= key.w[0xf];

  i = KEYCNT;
  j = 0;
  do
    {
      --i;
      u = h;
      h = g;
      g = f;
      f = e;
      e = d;
      d = c;
      c = b;
      b = a;
      a = u;
      u = key.t[j] + h;
      a -= key.t[i];
      a = BRTR1 (a, u >> 31);
      a = ROTR (a, u);
      a ^= g;
      a = BRTR4 (a, u >> 15);
      a -= f;
      a = BRTR3 (a, u >> 23);
      a = ROTR (a, u >> 5);
      a ^= e;
      a = BRTR2 (a, u >> 29);
      a -= d;
      a = BRTR4 (a, u >> 19);
      a = ROTR (a, u >> 10);
      a ^= c;
      a = BRTR3 (a, u >> 26);
      a -= b;
      a -= PW;
      ++j;
    }
  while (i != 0);

  a ^= key.w[0x0];
  b ^= key.w[0x1];
  c ^= key.w[0x2];
  d ^= key.w[0x3];
  e ^= key.w[0x4];
  f ^= key.w[0x5];
  g ^= key.w[0x6];
  h ^= key.w[0x7];
  SAVE (ip + 0x00, a);
  SAVE (ip + 0x04, b);
  SAVE (ip + 0x08, c);
  SAVE (ip + 0x0c, d);
  SAVE (ip + 0x10, e);
  SAVE (ip + 0x14, f);
  SAVE (ip + 0x18, g);
  SAVE (ip + 0x1c, h);
}



u32 __stdcall __export
getblocksize ()
{
  return 256;
}

u32 __stdcall __export
getkeysize ()
{
  return 512;
}

void __stdcall __export
getciphername (u8 * p)
{
  strcpy (p, "Paranoia");
}
